
import python_actr
import math

from .buffer import Chunk
import collections


class ImaginalModule(python_actr.Model):
  def __init__(self,buffer,delay=0.2,delay_sd=None):
    python_actr.Model.__init__(self)
    self._buffer=buffer
    self.busy=False
    self.delay=delay
    self.delay_sd=delay_sd

  def delay_time(self):
      t=self.delay
      if isinstance(t, collections.Callable): t=t()
      if self.delay_sd is not None:
          t=math.gauss(t,self.delay_sd)
      return t

  def set(self,chunk):
      if self.busy: return
      self.busy=True

      try:
          chunk=Chunk(chunk,self.sch.bound)
      except AttributeError:
          chunk=Chunk(chunk,{})        
      yield self.delay_time()

      self._buffer.chunk=chunk
      self.busy=False

  def modify(self,**args):
      if self.busy: return
      self.busy=True
      yield self.delay_time()
      self._buffer.modify(**args)
      self.busy=False

  def clear(self,**args):
      if self.busy: return
      self.busy=True
      yield self.delay_time()
      self._buffer.clear()
      self.busy=False

      
    





from . import logger

import os
import random
import re
import time

using_java=False
try:
    from java.io import File
    using_java=True
except:
    pass

def file_exists(filename):
    if hasattr(os,'access'):
        return os.access(filename,os.F_OK)
    elif using_java:
        return File(filename).exists()
        
    


def parse_code(lines):
    code=''
    defaults={}
    params=[]
    for i,line in enumerate(lines):
        line=line.strip()
        if len(line)==0: continue
        elif line.startswith('#'): continue
        elif '=' in line:
            k,v=line.split('=',2)
            params.append(k)
            defaults[k]=v
        elif '__future__' in line and 'generators' in line:
            continue
        else:
            code=''.join(lines[i:])
            break
    return params,defaults,code

def make_param_code(params,defaults,settings):
    p=[]
    for pp in params:
        v=defaults[pp]
        if pp in settings: v=fix_setting(settings[pp])
        p.append('%s=%s'%(pp,v))
    return '\n'.join(p)

def make_param_text(params,defaults,settings):
    p=[]
    for pp in params:
        if pp in settings and '%s'%settings[pp]!='%s'%defaults[pp]:
            v=settings[pp]
            p.append('%s(%s)'%(pp,v))
    if len(p)==0: return 'default'        
    return ' '.join(p)


def fix_setting(v):
    if not isinstance(v,(int,float)): 
        v=repr(v)
    return v


def make_settings_combinations(settings,keys=None):
    if keys is None: keys=list(settings.keys())
    if len(keys)==0: 
        yield {}
        return
    
    k=keys.pop()
    v=settings[k]
    for setting in make_settings_combinations(settings,keys):
        if type(v) is list:
            for vv in v:
                setting[k]=vv
                yield setting
        else:
            setting[k]=v
            yield setting

def ensure_backup(fn,lines):
    name=fn[:-3]+'/code.py'
    if not file_exists(fn[:-3]+'/'): os.makedirs(fn[:-3])
    if not file_exists(name):
        f=open(name,'w')
        f.write(''.join(lines))
        f.close()
    else:
        f=open(name,'r')
        if lines!=f.readlines():
            f.close()
            t=os.stat(name).st_mtime
            text=time.strftime('%Y%m%d-%H%M%S',time.localtime(t))
            name2='%s/code-%s.py'%(fn[:-3],text)
            
            os.rename(name,name2)
            f=open(name,'w')
            f.write(''.join(lines))
            f.close()
    

run_external=None
def run_with(script):
    global run_external
    run_external=script
    


def run(_filename,_iterations=1,**settings):
    if not _filename.endswith('.py'): _filename+='.py'
    if not file_exists(_filename):
        raise 'Could not find file: %s'%_filename

    lines=open(_filename,'r').readlines()
    params,defaults,core_code=parse_code(lines)
    
    ensure_backup(_filename,lines)
    
    
    f=None
    fname='.ccmtmp%08x.py'%random.randrange(0,0x70000000)

    for i in range(_iterations):
      for setting in make_settings_combinations(settings):
        param_code=make_param_code(params,defaults,setting)
        param_text=make_param_text(params,defaults,setting)
        
        if 'ccm.log' not in core_code:
            core_code='import ccm\nccm.log()\n'+core_code

        code='%s\n%s'%(param_code,core_code)
        code=code.replace('\r\n','\n')
        logline='ccm.log(data=True,screen=False,directory="%s/%s")'%(_filename[:-3],param_text)
        code=re.sub(r'ccm\.log\([^)]*\)',logline,code)
                
        print(_filename,'%d/%d'%(i,_iterations),param_text)
        
        f=open(fname,'w')
        f.write(code)
        f.flush()
        
        if run_external is None:
            compiled=compile(code,fname,'exec')
            exec(compiled, {})
            logger.finished()
        else:
            os.system('%s %s'%(run_external,fname))    
        f.close()
    os.remove(fname)
        


if __name__=='__main__':
    run()
  

        




import python_actr





def test_self():
    class ProdTest(python_actr.ProductionSystem):
      x=1
      def test(self='x:1'):
          self.x=2

    p=ProdTest()
    p.run()
    assert p.x == 2
    

def test_multi():
    class Prod(python_actr.ProductionSystem):
      class Module1(python_actr.Model):
         a=1
         def set(self,value):
             self.a=value
      class Module2(python_actr.Model):
         b=2
         def set(self,value):
             self.b=value
      state='swap'
      def swap(self='state:swap',Module1='a:?a',Module2='b:?b'):
          Module1.set(b)
          Module2.set(a)
          self.state=None

    p=Prod()
    p.run()
    assert p.Module1.a == '2'
    assert p.Module2.b == '1'

    
    
    
    


if __name__ == '__main__':
  unittest.main()     



from .buffer import Buffer, Chunk
from .dm import *
from .pm import *
from .compile import PMCompile

from .core import ACTR
from .timer import Timer
from .text import TextOutput
from .imaginal import ImaginalModule
from .vision import Vision
from .sosvision import SOSVision
from .motor import Motor



import python_actr

from python_actr.pattern import Pattern
from .dm import Finst
from .buffer import Chunk

class Vision(python_actr.Model):
  def __init__(self,visual,location):
    python_actr.Model.__init__(self)
    self._visual=visual
    self._location=location
    self.busy=False
    self.lastLocationPattern=Pattern('')
    self.tracking=None
    self.timeAppeared={}
    self.visualOnsetSpan=0.5
    self.finst=Finst(self)   
    
  def start(self):
    self.environmentUpdate()
    
  def isNew(self,object):
    if not getattr(object,'visible',True): return
    time=self.timeAppeared.get(object,None)
    if time==None:
      self.timeAppeared[object]=self.now()
      return True
    return self.now()<time+self.visualOnsetSpan
    
  def environmentUpdate(self):
   while True:    
    if self._location.isEmpty():
    
      r=[]
      for o in self.parent.parent.get_children():
        if not getattr(o,'visible',True):
            if o in list(self.timeAppeared.keys()): del self.timeAppeared[o] 
            continue
        if o not in self.timeAppeared:
          if hasattr(o,'x') and hasattr(o,'y'):  
            #if self.lastLocationPattern.match(o)!=None:
              r.append(o)
      if len(r)>0:
        obj=self.random.choice(r)    
        self.log._='Vision stuffed obj at (%g,%g)'%(obj.x,obj.y)
        self._location.set('%g %g'%(obj.x,obj.y))
    
    #print 'checking tracking',self.tracking  
    #if self.tracking!=None and (self.tracking not in self.parent.parent.get_children() or not getattr(self.tracking,'visible',True)):
    #  print '...lost'
    #  self.sch.add(self.lostTrack,delay=0.085)
    for o in self.parent.parent.get_children():    
      self.isNew(o)      
      
    yield self.parent.parent.changes,self.lostTrack
   
  def lostTrack(self):
    if self.tracking is not None:
      self.tracking=None
      self.log._='Object disappeared'
      self._visual.clear()
      #self._location.clear()

    
  def attendToUnattended(self,pattern=''):
    self.attendTo(pattern=pattern,unattended=True)
  def attendToNew(self,pattern=''):
    self.attendTo(pattern=pattern,new=True)

  def attendTo(self,pattern,unattended=False,new=False):
    if isinstance(pattern, str) and ':' not in pattern and pattern.count(' ')==1:
        pattern='x:%s y:%s'%tuple(pattern.split(' '))
  
    self.lastLocationPattern=Pattern(pattern,self.sch.bound)
    r=[]
    for obj in self.parent.parent.get_children():
      if new==True and not self.isNew(obj): continue
      if unattended==True and self.finst.isIn(obj): continue
      if hasattr(obj,'x') and hasattr(obj,'y') and getattr(obj,'visible',True):
        if self.lastLocationPattern.match(obj)!=None:
          r.append(obj)
    if len(r)>0:
      obj=self.random.choice(r)    
      self.log._='Vision found obj at (%g,%g)'%(obj.x,obj.y)
      self._location.set('%g %g'%(obj.x,obj.y))

  def isClose(self,a,b):
      return abs(a-b)<0.01

  def examine(self,pat):
    if isinstance(pat, str) and ':' not in pat and pat.count(' ')==1:
        pat='x:%s y:%s'%tuple(pat.split(' '))
    
    self.lastExamine=Pattern(pat,self.sch.bound)
    
    pat=self.lastExamine

    for obj in self.parent.parent.get_children():
     if hasattr(obj,'x') and hasattr(obj,'y') and getattr(obj,'visible',True):   
      if pat.match(obj) is not None:
      #if self.isClose(obj.x,float(pat[0])) and self.isClose(obj.y,float(pat[1])):        
        self.busy=True
        yield 0.085
        self.busy=False
        if obj in self.parent.parent.get_children():
          self.tracking=obj
          self.log._='Vision sees %s'%Chunk(obj)
          self._visual.set(obj)
          self.finst.add(obj)
        else:
          self.tracking=None  
          self.log._='Vision sees nothing'
          self._visual.clear()
        break  




import unittest

from python_actr import *

class MyModel(ACTR):
  focus=Buffer()
  retrieve=Buffer()
  dm=Memory(retrieve)
  pgc=PMPGC()
  compiler=PMCompile(keep='focus',request='dm.request',retrieve='retrieve')
  
  def init():
    dm.add('pattern a 7')
    dm.add('pattern b 3')
 
  def prod1(focus='see ?x'):
    dm.request('pattern ?x ?')
    focus.set('remember ?x')
 
  def prod2(focus='remember ?x',retrieve='pattern ?x ?y'):
    focus.set('say ?y')
    self.success()
    self.stop()
    
    
class TestScheduler(unittest.TestCase):
  def setUp(self):
    pass

  def test_compile(self):
        m=MyModel()

        m.focus.set('see a')
        m.run()
        self.assertEqual(str(m.focus.chunk),'say 7')
        self.assertAlmostEqual(m.now(),0.15)
    
        m.focus.set('see a')
        m.run()
        self.assertEqual(str(m.focus.chunk),'say 7')
        self.assertAlmostEqual(m.now(),0.20)

        m.focus.set('see b')
        m.run()
        self.assertEqual(str(m.focus.chunk),'say 3')
        self.assertAlmostEqual(m.now(),0.35)
    
        m.focus.set('see b')
        m.run()
        self.assertEqual(str(m.focus.chunk),'say 3')
        self.assertAlmostEqual(m.now(),0.40)

    
if __name__ == '__main__':
  unittest.main()     





from . import scheduler            
from . import logger
import random
import inspect      
import copy
#import ccm.config as config

class MethodWrapper:
  def __init__(self,obj,func,name):
      self.func=func
      self.obj=obj
      self.__name__=name
      self.begins=scheduler.Trigger(name+' begin')
      self.ends=scheduler.Trigger(name+' end')
      self.default_trigger=self.ends
  def __call__(self,*args,**keys):
      self.obj.sch.trigger(self.begins)
      val=self.func(self.obj,*args,**keys)
      self.obj.sch.trigger(self.ends)
      return val
class MethodGeneratorWrapper(MethodWrapper):
  def _generator(self,*args,**keys):
      self.obj.sch.trigger(self.begins)
      for x in self.func(self.obj,*args,**keys):
          yield x
      self.obj.sch.trigger(self.ends)
  def __call__(self,*args,**keys):
      return self.obj.sch.add(self._generator,args=args,keys=keys)
  def __str__(self):
      return '<MGW %s %s>'%(self.obj,self.__name__)    

def log_everything(model,log=None):
  if log is None: log=logger.log_proxy
  if not hasattr(model,'log'): model.run(limit=0)
  model.log=log
  for k,v in list(model.__dict__.items()):
    if k[0]!='_' and k!='parent':
      
      if isinstance(v,Model) and v.parent is model:
        log_everything(v,getattr(log,k))


class Model:
    __converted=False
    _convert_methods=True
    _auto_run_start=True
    name='top'
    


    def __init__(self,log=None,**keys):
        #print("this happens")
        #print(self,"Model A1")#sterling
        self.__init_log=log
        for k,v in list(keys.items()):
          setattr(self,k,v)

    #getitem added by sterling
    def __getitem__(self,item):
        return getattr(self,item)



    def __convert(self,parent=None,name=None):
        #print("__convert() called", self)
        #Culprit
        #if self.__converted: return
        assert self.__converted==False
        self.__converted=True    
        self.changes=scheduler.Trigger()
        
        if hasattr(self,'parent'): parent=self.parent
        
        methods={}
        objects={}
        #print(inspect.getmro(self.__class__),"MRO") #sterling
        #This MRO includes class object, whereas in 2.7, it does not
        #add[:-1], because it seems that object is always at the end.
        '''            obj.__convert(self,name)
                    print("run after")'''#(from line 132) now works.
        #BUT a later run crashes.
        for klass in inspect.getmro(self.__class__)[:-1]:
          
            #print(klass, "klass")
            if klass is not Model:
                #print(klass, "klass")
                for k,v in inspect.getmembers(klass):
                    #print(k,",",v,"k,v")
                    
                    if k[0]!='_':
                        #print(k,"just k")
                        #A problem. ismethod(v) is not returning true.
                        #Did run an exmample, and it worked fine.
                        #It is working though, because they are unbound methods, and ismethod
                        #should only return true if they are bound
                        #isfunction might work in it's place
                        #if inspect.ismethod(v):
                        if inspect.isfunction(v):
                            #print(k,v,"k,v+")
                            if k not in ['run','now','get_children'] and k not in methods and klass is not Model:
                                #print(k,v,"k,v-post")
                                methods[k]=v
                        else:
                            if inspect.isclass(v) and Model in inspect.getmro(v):
                                v=v()
                            if k not in objects:
                                objects[k]=v
        objects=copy.deepcopy(objects)
        #works up to here
        
        if parent:
            if not parent.__converted: parent.__convert()
            self.sch=parent.sch
            self.log=logger.dummy
            self.random=parent.random
            self.parent=parent
        else:
            self.sch=scheduler.Scheduler()
            if self.__init_log is True:
                self.log=logger.log()
            elif self.__init_log is None:
                self.log=logger.dummy   
            else:
                self.log=self.__init_log    
            self.random=random.Random() 
            #seed=config.getOptions().random
            #if seed is not None: self.random.seed(seed)
            
            self.parent=None   
        
        self._convert_info(objects,methods)
        #works up to here
        #print(list(objects.items()))
        for name,obj in list(objects.items()):
            #print("run before before",name,obj)
            #self.run(limit=0)#print(name,obj,"name,obj")
            if isinstance(obj,Model):
                #print(dir(obj))  
                if not obj.__converted:
                    #print(obj,"obj")
                    #print(self,"self")
                    #print("run before")
                    #self.run(limit=0)
                    
                    #self.run(limit=0)
                    obj.__convert(self,name)
                    #print("run after")
                    
                    #self.run(limit=0)
                else:
                    obj.name=name    
                try:
                  self._children[name]=obj
                except AttributeError:
                  self._children={name:obj}
            #print("run before 2",name,obj)
            #self.run(limit=0)
            #print(dir(obj))
            self.__dict__[name]=obj
            #print("run after 2",name,obj)
            #print(dir(obj))
            #self.run(limit=0)
        #check out _convert_info - see difference between 2.7 running and 3.3 running.
        #does not between here and last
        
        if self._convert_methods:    
          for name,func in list(methods.items()):
              #print("here",name,func)
              #print(dir(func))
              #if func.__func__.__code__.co_flags&0x20==0x20:
              if func.__code__.co_flags&0x20==0x20:
                  w=MethodGeneratorWrapper(self,func,name)
              else:
                  w=MethodWrapper(self,func,name)
              self.__dict__[name]=w    

        if self._auto_run_start:
            self.start()
        
        for k,v in list(self.__dict__.items()):
            if k[0]!='_' and k!='parent' and isinstance(v,Model):
                if not v.__converted:
                    v.__convert(parent=self)
                
        
    def _convert_info(self,objects,methods):
        pass    
                
    def __setattr__(self,key,value):
      if key=='parent' and value is None and getattr(self,'parent',None) is not None:
        del self.parent._children[self.name]
      self.__dict__[key]=value
      

      if isinstance(value,Model) and key[0]!='_' and key!='parent':
        self._ensure_converted()
        p=self
        ancestor=True
        while p is not None:
          if value is p: break
          p=getattr(p,'parent',None)
        else:
          if getattr(value,'parent',None) is not None:
            pass
            #if value.name in value.parent._children:
            #  del value.parent._children[value.name]
          else:  
            value.parent=self
            value.name=key
            try:
              self._children[key]=value
            except AttributeError:
              self._children={key:value}  
            if self.__converted and not value.__converted: value.__convert(name=key,parent=self)
              
      if self.__converted and key[0]!='_' and key not in ['parent','sch','changes','log','random','name']:
          m=self
          done=[]
          while m is not None:
            self.sch.trigger(m.changes,priority=-1)
            done.append(m)
            m=m.parent
            if m in done: m=None
          if self.log: setattr(self.log,key,value)
      
      if key=='log' and value is not None:
        for k,v in list(self.__dict__.items()):
          if k[0]!='_' and k not in ['parent','sch','changes','log','random','name']:
            if isinstance(v,(int,str,float,type(None))):
              setattr(value,k,v)
          
    
    
    
    def start(self):
        pass
    def run(self,limit=None,func=None):
        if not self.__converted:
            self.__convert()
        #if config.getOptions().logall: log_everything(self)
        if limit is not None: 
            self.sch.add(self.sch.stop,limit,priority=-9999999)
        if func is not None:
            self.sch.add(func)
        self.sch.run()
    def stop(self):
        if not self.__converted:
            self.__convert()
        self.sch.stop()
    def now(self):
        if not self.__converted:
            self.__convert()
        return self.sch.time
    
    def get_children(self):
        try:
          return list(self._children.values())
        except AttributeError:
          return []  
        
        
    def _get_scheduler(self):
        self._ensure_converted()
        return self.sch

    def _ensure_converted(self):
        if not self.__converted:
            self.__convert()

    def _is_converted(self):
        return self.__converted
        
                    
      
      
      
      
      
    







from .model import Model, log_everything
from .production import ProductionSystem
from .logger import log, finished
from .runner import run, run_with
from .display import display
from .actr import *
from . import version



import atexit
import time
import bisect
import sys
import random
import os

using_java=False
try:
    from java.io import File
    using_java=True
except:
    pass

def file_exists(filename):
    if hasattr(os,'access'):
        return os.access(filename,os.F_OK)
    elif using_java:
        return File(filename).exists()


class Trace:
    def __init__(self):
        self.data=dict(time=[(0,0.0)])
        self.index=1
        self.type={}
    def add(self,key,value):
        if not isinstance(value,(bool,int,float,str,type(None))):
            value=repr(value)
        if type(value)==str:
            if value.startswith('<'): return
        if key not in self.data:
            self.data[key]=[]

        self.data[key].append((self.index,value))
        self.index+=1
            
    def keys(self):
        return list(self.data.keys())
    def fixed_keys(self):
        return [k for k,v in list(self.data.items()) if len(v)==1]

    def get_final(self,key):
        return self.data[key][-1][1]
    def get_pts(self,vars):
        pts=[]
        for v in vars:
            for i,val in self.data[v]:
                if i not in pts: pts.append(i)
        pts.sort()
        return pts    

    def merge_pts(self,pts,key):
        times=[x[0] for x in self.data[key]]
        i=0
        for t in times:
            if i>=len(pts) or t<pts[i]: continue
            while i<len(pts)-1 and pts[i+1]<t:
                del pts[i]
            i+=1
        while i<len(pts)-1:
            del pts[i]

    def group_pts(self,pts,key):
        times=[x[0] for x in self.data[key]]
        group=[]
        i=0
        for p in pts:
            if i>=len(times) or p<times[i]: group.append(p)
            else:
                if len(group)>0: yield group
                group=[p]
                while i<len(times) and p>=times[i]:
                    i+=1
        if len(group)>0: yield group  

    def get_at(self,name,time):
        d=self.data[name]
        i=bisect.bisect(d,(time+0.5,))-1
        if i<0: return None
        if i>=len(d)-1: i=-1
        return d[i][1]

    def __bool__(self):
        return self.index!=1



class Log:
    def __init__(self):
        self.do_screen=True
        self.do_html=False
        self.do_summary=False
        self.do_data=False
        self.directory=sys.argv[0]
        self.using_default_directory=True
        if self.directory.endswith('.py'): self.directory=self.directory[:-3]
        
        self.start_time=time.time()
        self.last_flush=self.start_time
        
        self.reset()
        
    def use_directory(self,dir):
        self.directory=dir
        self.using_default_directory=False    
        

    def set(self,key,value):
        if key=='time': self.time=value
        if self.do_screen:
            self.display_value(key,value)
        if self.do_data or self.do_summary:
            self.data[key]=value
        if self.do_html:    
            self.trace.add(key,value)

    def display_value(self,key,value):
        if key!='time':
            print('%8.3f %s %s'%(self.time,key,value))
    def display_all(self):
        if self.time>0:
            print('Total time: %8.3f'%self.time)
        items=list(self.data.items())
        items.sort()
        for k,v in items:
            if k!='time':
                print(' %s %s'%(k,v))

    def get_time_code(self):
        t=time.strftime('%Y%m%d-%H%M%S',time.localtime(self.start_time))
        return '%s-%s'%(t,self.id)
        
    def ensure_directory_exists(self):
        if not file_exists(self.directory+'/'): os.makedirs(self.directory)
        
    def reset(self):
        self.trace=Trace()
        self.data={}
        self.time=0
        self.id='%08x'%int(random.randrange(0x7FFFFFFF))
  


class DummyLog:
  def set(self,key,value):
    pass
  def __bool__(self):
    return False
  def __setattr__(self,key,value):
    pass
  def __getattr__(self,key):
    if key[0]!='_': return self
    else: raise AttributeError(key)
  def __setitem__(self,key,value):
    pass
  def __getitem__(self,key):
    return self
dummy=DummyLog()    


class LogProxy:
  def __init__(self,log,prefix=''):
    self._log=log
    if prefix.startswith('.'): prefix=prefix[1:]
    self._prefix=prefix
    self._sub={}
  def __setattr__(self,key,value):
    if key.startswith('_'):
      if key=='_': self._set(value)
      else: self.__dict__[key]=value
    else:
      getattr(self,key)._set(value)  

  def __getattr__(self,key):
    if key.startswith('_'):
        if key=='_': return self
        if key in self.__dict__: return self.__dict__[key]
        raise AttributeError(key)
    else:
      s=LogProxy(self._log,'%s.%s'%(self._prefix,key))
      self.__dict__[key]=s
      return s
        
  def __setitem__(self,key,value):
      self[key]._set(value)  
  
  def __getitem__(self,key):
      s=self._sub.get(key,None)
      if s is not None: return s
      s=LogProxy(self._log,'%s[%s]'%(self._prefix,key))
      self._sub[key]=s
      return s

  def _set(self,value):
    try:  
      if not isinstance(value,(int,float,bool,str,tuple,list,dict)):
        value=repr(value)
    except TypeError:
        value=repr(value)
    self._log.set(self._prefix,value)  
  
  def __bool__(self):
    return True




singleton_log=Log()
log_proxy=LogProxy(singleton_log)
def log(screen=None,html=None,data=None,summary=None,directory=None):
    if screen is not None: singleton_log.do_screen=screen
    if html is not None: singleton_log.do_html=html
    if data is not None: singleton_log.do_data=data
    if summary is not None: singleton_log.do_summary=summary
    if directory is not None: singleton_log.use_directory(directory)
    return log_proxy


pending_output=[]
def finished(flush=True):
    
    log=singleton_log
    has_data=log.data or log.trace
    
    if has_data:
        
        if log.do_summary:
            log.display_all()
        if log.do_data:
            log.ensure_directory_exists()
            fn='%s/%s.data'%(log.directory,log.get_time_code())
            pending_output.append((fn,log.data))
        if log.do_html:
            log.ensure_directory_exists()
            from .ui.htmltrace import HTMLTrace
            html=HTMLTrace(log.trace)
            html.generate('%s/%s.html'%(log.directory,log.get_time_code()))

    if flush or time.time()-log.last_flush>10:
        for fn,data in pending_output:
            f=open(fn,'w')
            items=list(data.items())
            items.sort()
            for k,v in items:
                f.write('%s=%s\n'%(k,v))
            f.close()
        del pending_output[:]
        log.last_flush=time.time()        
    log.reset()
    print ("end...")
    
    

atexit.register(finished,flush=True)   




import unittest

import python_actr
from python_actr import *


class Basic(ACTR):
    goal=Buffer()
    retrieval=Buffer()
    memory=Memory(retrieval)
    salience=DMSalience(memory)
    
    def init():
        goal.set('test')
        memory.add('id:1 color:blue shape:X')
        memory.add('id:2 color:blue shape:X')
        memory.add('id:3 color:blue shape:X')
        memory.add('id:4 color:blue shape:X')
        memory.add('id:5 color:blue shape:O')
        memory.add('id:6 color:red shape:X')
        salience.context('')
        
    def test(goal='test'):
        memory.request('')
        goal.set('test2')
    
    def test2(goal='test2',retrieval='id:?id color:?color shape:?shape'):
        goal.set('id:?id color:?color shape:?shape')
        self.stop()



class TestACTRMemory(unittest.TestCase):
    def test_basic(self):
        p=Basic()
        p.salience.weights(color=1)
        p.run()
        self.assertEqual(p.goal['id'],'6')
            
        p=Basic()
        p.salience.weights(shape=1)
        p.run()
        self.assertEqual(p.goal['id'],'5')


if __name__ == '__main__':
  unittest.main()     



import re
import typing

class PatternException(Exception):
    pass

def get(obj,name,key):    
  if name is None: a=obj
  else:
      a=obj[name]

  while type(key)==str and '.' in key:
      key1,key=key.split('.',1)
      try:
        a=a[key1]
      except AttributeError:
        a=getattr(a,key1)

  try:
      x=a[key]
  except AttributeError:    
      x=getattr(a,key)
  if isinstance(x,float): x='%g'%x    
  if not isinstance(x,str): x=repr(x)
  return x

def partialmatch(obj,name,key,b,value):
  if type(key)==str and key[0]=='?':
      key=b[key[1:]]
  v=get(obj,name,key)
  if v==value: return True

  # fix for early Python versions where True and False are actually 1 and 0
  if value in ['True','False'] and type(True)==int:
      if v==str(bool(value)): return True
      
  pm=b.get('_partial',None)
  if pm is not None:
    x=pm.match(key,value,v)
    obj._partial+=x
    return True
  else:
    return False
    


class Pattern:
    def __init__(self,patterns,bound=None,partial=None):
        self.funcs=parse(patterns,bound)
        self.partial=partial
        
    def match(self,obj):
        #print("-OBJ",dir(obj))#sterling
        b={}
        b['_partial']=self.partial
        if self.partial is not None:
          obj._partial=0.0

        #import traceback#sterling
        try:#sterling, removed try
            for f in self.funcs:

        #       #print("f",dir(f), repr(f))#sterling
                if f(obj,b)==False:
        #            print("RETURNING NONE?????")#sterling
                    return None
        except (AttributeError,TypeError,KeyError) as e:
            #print("EXCEPTION HAPPENEND",dir(e)) #sterling
            #traceback.print_exc()
            return None
        del b['_partial']
        return b    
            


        
        
def parse(patterns,bound=None):
    if not hasattr(patterns,'items'):
      patterns={None:patterns} 
    funcs=[]
    vars={}
    funcs2=[]
    #print("parse patterns", patterns)#sterling
    for name,pattern in list(patterns.items()):
        #print("name,patter",name,pattern)#sterling
        if not isinstance(pattern,(list,tuple)): pattern=[pattern]

        for p in pattern:
            #print("p in pattern", p)#sterling
            if p is None:
                if name is None: funcs.append(lambda x,b: x==None)
                else:            funcs.append(lambda x,b,name=name: x[name]==None or len(x[name])==0)
            elif isinstance(p, typing.Callable):
                if name is None:
                  def callfunc(x,b,name=name,p=p):
                    return p(x,b)
                else:
                  def callfunc(x,b,name=name,p=p):
                    return p(x[name],b)
                funcs2.append(callfunc)        
            elif isinstance(p,str):
                #print("p is a string!")#sterling
                namedSlots=False
                for j,text in enumerate(p.split()):
                    key=j
                    m=re.match('([?]?[\w\.]+):',text)
                    #print("1m", m)#sterling
                    if m!=None:
                        key=m.group(1)
                        try:
                            key=int(key)
                        except ValueError:
                            pass
                        text=text[m.end():]
                        #print("text", text)#sterling
                        if len(text)==0:
                            raise PatternException("No value for slot '%s' in pattern '%s'"%(key,pattern))
                        namedSlots=True          
                    else:
                        if namedSlots!=False:
                            raise PatternException("Found unnamed slot '%s' after named slot in pattern '%s'"%(text,pattern))
                    if text=='?': continue
                    while len(text)>0:
                        m=re.match('([\w\.-]+)',text)
                        #print("2m", m)#sterling
                        if m!=None:
                            text=text[m.end():]
                            t=m.group(1)
                            #print("name,key,t",name,key,t)#sterling
                            funcs.append(lambda x,b,name=name,key=key,t=t: partialmatch(x,name,key,b,t))
                            continue

                        m=re.match('!([\w\.-]+)',text)
                        #print("3m", m)#sterling
                        if m!=None:
                            text=text[m.end():]
                            t=m.group(1)
                            funcs.append(lambda x,b,name=name,key=key,t=t: get(x,name,key)!=t)
                            continue
        
                        m=re.match('\?(\w+)',text)
                        #print("4m", m)#sterling
                        if m!=None:
                            text=text[m.end():]
                            v=m.group(1)
                            if bound is not None and v in bound:
                                funcs.append(lambda x,b,name=name,key=key,t=bound[v]: partialmatch(x,name,key,b,t))
                            elif v in vars:
                                funcs2.append(lambda x,b,name=name,key=key,v=v: partialmatch(x,name,key,b,b[v]))
                            else:    
                                vars[v]=(name,key)
                                def setfunc(x,b,name=name,key=key,v=v):
                                  b[v]=get(x,name,key)
                                  return True
                                funcs.append(setfunc)
                            continue
        
                        m=re.match('!\?(\w+)',text)
                        #print("5m", m)#sterling
                        if m!=None:
                            text=text[m.end():]
                            v=m.group(1)
                            if bound is not None and v in bound:
                                funcs.append(lambda x,b,name=name,key=key,t=bound[v]: get(x,name,key)!=t)
                            else:
                                funcs2.append(lambda x,b,name=name,key=key,v=v: get(x,name,key)!=b[v])
                            continue

                        raise PatternException("Unknown text '%s' in pattern '%s'"%(text,pattern))  
    return funcs+funcs2                        
        

        
        
        
        




import python_actr
import typing

class ACTR(python_actr.ProductionSystem):
    production_time_sd=None
    production_threshold=None
    
    def __init__(self,log=None):
        #ccm.ProductionSystem.__init__(self,log=log)
        #print(self,"ACTR")
        super(ACTR,self).__init__(log=log)
        self._adaptors=[]
    
    def _process_productions(self):
        self._calc_context()          
        for i in self._initializers:
            i.fire(self._context)
        while True:
            if self.production_match_delay>0: yield self.production_match_delay 
            match=[p for p in self._productions if p.match(self._context)]
            if len(match)==0:
                yield self._top.changes
            else:
                activations=[self.get_activation(p) for p in match]
                a=max(activations)
                
                threshold=self.production_threshold
                if isinstance(threshold, typing.Callable): threshold=threshold()
                
                if threshold is not None and a<threshold:
                    for a in self._adaptors: a.below_threshold()
                    yield self._top.changes
                    continue    
                options=[p for (i,p) in enumerate(match) if activations[i]==a]
                choice=self.random.choice(options)
                
                for a in self._adaptors: a.selecting(choice)
                #self.log.selected=choice.name
                self.log.production=choice.name
                
                t=self.production_time
                if isinstance(t, typing.Callable): t=t()
                if self.production_time_sd is not None:
                    t=t+self.random.gauss(0,self.production_time_sd)
                t-=self.production_match_delay
                if t<0: t=0
                yield t
                
                if not choice.match(self._context):
                  #self.log.change_detected='before firing '+choice.name
                  self.log.production='(changed before firing)'
                else:  
                  for a in self._adaptors: a.firing(choice)
                  self.log.production=None
                  #self.log.firing=choice.name
                  choice.fire(self._context)
                  yield dict(delay=0,priority=-1000)      # delay so we don't try to match again until after the result of production firing has had a chance to occur
    
    def add_adaptor(self,module):
        self._adaptors.append(module)
        for p in self._productions:
            module.create(p)
            
    def reward(self,value):
        for a in self._adaptors:
            a.reward(value)
    def success(self):
        self.reward(1)
    def failure(self):
        self.reward(-1)            
                
                

    def get_activation(self,production=None):
        if production is None:
            r={}
            for p in self._productions:
                r[p.name]=self.get_activation(p)
            return r
        elif isinstance(production,str):
            for p in self._productions:
                if p.name==production:
                    return self.get_activation(p)
            return None
        else:
            activation=production.base_utility
            for a in self._adaptors:
                activation+=a.utility(production)
            return activation
            
       



from .pytag import *

def splitKey(key):
  r=[]
  a=''
  depth=0
  for c in key:
    if c=='.' and depth==0: 
      if a:
        r.append(a)
        a=''
    elif c in '[(':
      if a:
        r.append(a)
        a=''
      depth+=1
    elif c in '])':
      if a:
        r.append(a)
        a=''
      depth-=1
    else:
      a+=c  
  if a:
    r.append(a)
    a=''
  return r    
      
      
    

  return key.split('.')

def makeHeader(table,keys):
  keys=[splitKey(k) for k in keys]
  size=max(len(x) for x in keys)
  for k in keys:
    while len(k)<size: k.append('')
    
  noMerge=[False]*len(keys)
  for i in range(size):
    row=[keys[j][i] for j in range(len(keys))]
    merged=[]
    values=[row[0]]
    count=1
    for j in range(1,len(keys)):
      if noMerge[j] or row[j]!=values[-1]:
        merged.append(count)
        count=1
        values.append(row[j])
        noMerge[j]=True
      else:
        count+=1
    merged.append(count)
    
    row=tr()
    for j in range(len(merged)):
      row[th(colspan=repr(merged[j]))[values[j]]]
    table[row]  
          
      
colors="""AliceBlue
AntiqueWhite
Aqua  
Aquamarine
Azure  
Beige  
Bisque 
BlanchedAlmond 
BurlyWood 
Chartreuse 
Cornsilk 
Cyan 
DarkGrey 
DarkKhaki 
Darkorange 
DarkSalmon  	
DarkSeaGreen 
DarkTurquoise 
DeepSkyBlue 
DodgerBlue 
Gainsboro 
GhostWhite
Gold  
GoldenRod 
GreenYellow
HoneyDew 
Ivory  	
Khaki  	
Lavender 
LavenderBlush
LawnGreen
LemonChiffon
LightBlue 
LightCyan  	
LightGoldenRodYellow
LightGray  
LightGrey  
LightGreen 
LightPink  
LightSeaGreen 
LightSkyBlue 
LightSteelBlue 
LightYellow  
Lime
LimeGreen
Linen 
MediumAquaMarine 
MediumSeaGreen 
MediumSpringGreen
MediumTurquoise 
MintCream  	
MistyRose  
Moccasin  	
NavajoWhite 
OldLace 
Orange 
PaleGoldenRod 
PaleGreen  
PaleTurquoise 
PapayaWhip
PeachPuff 
Pink  
Plum  
PowderBlue 
Salmon  
SandyBrown 
Silver  
SkyBlue 
SpringGreen 
Tan 
Thistle 
Turquoise 
Wheat 
WhiteSmoke  
Yellow  
YellowGreen""".split()
    
    
    
    
    
  


class HTMLTrace:
  def __init__(self,trace):
    self.trace=trace
    
    
  def getColor(self,value):
    if value=='': return 'white','white'
    if value=='True' or value is True: return 'lightgreen','green'
    if value=='False' or value is False: return 'pink','red'
    if isinstance(value,(int,float)): return 'black','white'
    num=hash(value)
    return 'black',colors[num%len(colors)]

  def fixValue(self,val): 
    if val is None or val=='None': val=''
    try:    val=val.replace('<','&lt;').replace('>','&gt;')  
    except: pass  
    if type(val) not in [int,float,bool] and ':' in val:
      slots=val.split()
      for i,slot in enumerate(slots):
        if ':' in slot:
          a,b=slot.split(':',1)
          slots[i]='<i>%s:</i>%s'%(a,b)
      val=' '.join(slots)    
    
    return val  


  def makeFixedTable(self,fixed):
    t=table()
    for k in fixed:
      t[tr[td[k],td[self.trace.get_final(k)]]]
    return t
      
    
  def makeBody(self,table,keys,pts):
   
    grouped={}
    for k in keys:
      grouped[k]=list(self.trace.group_pts(pts,k))
    
    for pt in pts:
      row=tr()
      for k in keys:
        if pt not in grouped[k][0]:
          del grouped[k][0]
          
        if pt==grouped[k][0][0]:
          val=self.trace.get_at(k,pt)
          val=self.fixValue(val)
          if k=='time':
              val='%1.3f'%val
              c,bg='white','#333333'
          else:
              c,bg=self.getColor(val)

          style='background:%s; color:%s;'%(bg,c)
          row[td(rowspan=repr(len(grouped[k][0])),style=style)[val]]
      table[row]  
        
    
  def generate(self,filename):
    keys=list(self.trace.keys())
    fixed_keys=self.trace.fixed_keys()
    fixed_keys.sort()
    keys=[k for k in keys if k not in fixed_keys]
    keys.sort()
    
    has_time=False
    if 'time' in keys:
      keys.remove('time')
      has_time=True
    pts=self.trace.get_pts(keys)
    if has_time:
      keys.insert(0,'time')
      timePts=self.trace.get_pts(['time'])
    
    if 'time' in keys:
      self.trace.merge_pts(pts,'time')
      

         
    
    tbl=table()
    makeHeader(tbl,keys)
    self.makeBody(tbl,keys,pts)


    fixed=self.makeFixedTable(fixed_keys)    
    
    if not filename.endswith('.html'): filename+='.html'    
    f=open(filename,'w')
    
    page=html[
           head[
             title[filename],
             style["""
                     table {border-collapse: collapse; empty-cells:show;}
                     td {border: solid black 1px; vertical-align:top;}
                     th {border: solid #cccccc 1px; background:black; color:white;}
                   """],
             ],
           body[
             tbl,
             fixed,
             ]
           ]  
    
    print(page, file=f)
    
    



engines=dict(tk='python_actr.display.tk.core.TkinterDisplay',
             cairo='python_actr.display.cairo.core.CairoDisplay',
             pygame='python_actr.display.pygame.core.PygameDisplay',
             )

default_order=['tk','pygame']


def display(root,engine=None,**args):
    d=None
    error=""

    order=default_order[:]
    if engine is not None and engine in order:
        order=[engine]
    
    for e in order:
        module,obj=engines[e].rsplit('.',1)
        try:
            m=__import__(module,globals(),locals(),[obj])
            d=getattr(m,obj)(root,**args)
            return d
        except ImportError as e:
            error+='\n'+str(engines[e])
        
    print('Error: could not create display: %s'%error)



import unittest

import python_actr
from python_actr import *


class Basic(ACTR):
    goal=Buffer()
    retrieval=Buffer()
    memory=Memory(retrieval)
    
    def init():
        goal.set('test')
        memory.add('hello world')
        
    def test(goal='test'):
        memory.request('hello ?')
        goal.set('test2')
    
    def test2(goal='test2',retrieval='hello ?x'):
        goal.set('?x')    
        self.stop()


class TwoWords(ACTR):
    goal=Buffer()
    retrieval=Buffer()
    memory=Memory(retrieval)
    
    def init():
        goal.set('test')
        memory.add('hello world')
        memory.add('bye world')
        
    def test(goal='test'):
        memory.request('? ?')
        goal.set('test2')
    
    def test2(goal='test2',retrieval='?a ?x'):
        goal.set('?x')    
        self.stop()
        

class TestACTRMemory(unittest.TestCase):
    def test_basic(self):
        p=Basic()
        #ccm.log_everything(p)
        p.run()
        self.assertEqual(p.goal[0],'world')
        self.assertAlmostEqual(p.now(),0.15)
            
    def test_two(self):
        p=TwoWords()
        #ccm.log_everything(p)
        p.run()
        self.assertEqual(p.goal[0],'world')
        self.assertAlmostEqual(p.now(),0.15)
        
    def test_lambda(self):
        class LambdaTest(ACTR):
            focal=Buffer()
            retrieval=Buffer()
            memory=Memory(retrieval)
            def init():
                focal.set('test')
                memory.add('x:30 y:50 text:A')
                memory.add('x:80 y:50 text:B')
            def test(focal='test'):
                memory.request(['y:50',lambda obj,b:int(obj['x'])<50])
                focal.set('test2')
            def test2(focal='test2',retrieval='text:?text'):
                focal.set('?text')
                self.stop()
        p=LambdaTest()
        p.run()
        self.assertEqual(p.focal.chunk[0],'A')


if __name__ == '__main__':
  unittest.main()     




import python_actr
import math

__all__=['Memory','MemorySubModule','DMNoise','DMBaseLevel','DMSalience','DMSpreading','DMFixed',
         'Partial','BlendingMemory','DMAssociate','DMInhibition']

from .buffer import Chunk,Buffer
from python_actr.pattern import Pattern

class Memory(python_actr.Model):
  def __init__(self,buffer,latency=0.05,threshold=0,maximum_time=10.0,finst_size=4,finst_time=3.0):
    python_actr.Model.__init__(self)
    self._buffer=buffer
    self.dm=[]
    self.error=False
    self.busy=False
    self.adaptors=[]
    self.latency=latency
    self.threshold=threshold
    self.maximum_time=maximum_time
    self.partials=[]
    self.finst=Finst(self,size=finst_size,time=finst_time)
    self.record_all_chunks=False
    self._request_count=0
    
  def clear(self):
    del self.dm[:]
    
  def add(self,chunk,record=None,**keys):
    if self.error: self.error=False
    if isinstance(chunk,Buffer):
      chunk=Chunk(chunk.chunk)
    else:
      bound=None
      if hasattr(self,'sch'):
        bound=getattr(self.sch,'bound',None)
      chunk=Chunk(chunk,bound)
    for c in self.dm:
      if chunk==c:
        for a in self.adaptors: a.merge(c,**keys)
        break
    else:
      for a in self.adaptors: a.create(chunk,**keys)
      self.dm.append(chunk)
    chunk.record=record

      
  def find_matching_chunks(self,pattern,threshold=None):
     bound=getattr(self.sch,'bound',None)
     pattern=Pattern(pattern,bound)
     matches=[x for x in self.dm if pattern.match(x) is not None]
     if threshold is not None:
       matches=[x for x in matches if self.get_activation(x)>=threshold]
     return matches  
        
      
  def request(self,pattern,partial=None,require_new=False):
     if partial is None and len(self.partials)>0: partial=self.partials[0]
     self.busy=True
     if self.error: self.error=False
     self._request_count+=1
     b=getattr(self.sch,'bound',None)
     pattern=Pattern(pattern,b,partial=partial)
     
     all=self.dm
     if require_new: all=[x for x in all if not self.finst.contains(x)]
     
     matches=[x for x in all if pattern.match(x) is not None]
     
     #for x in matches: print `x`,self.get_activation(x)
     
       
     for a in self.adaptors: a.matched(matches)
     if len(matches)==0:
         self.fail(self._request_count)
     else:
         maximum=None    
         for chunk in matches:
             chunk.activation=self.get_activation(chunk)
             if partial is not None:
               chunk.activation+=chunk._partial
             if maximum is None or chunk.activation>maximum:
               maximum=chunk.activation
         if maximum<self.threshold:
             self.fail(self._request_count)
         else:    
             best=[x for x in matches if x.activation==maximum]
             
             choice=self.random.choice(best)
             self.recall(choice,matches=matches,request_number=self._request_count)
     
  def fail(self,request_number):
     if self.threshold is None: 
         time=self.maximum_time
     else:
         time=self.latency*math.exp(-self.threshold)
         if time>self.maximum_time: time=self.maximum_time 
     yield time
     if request_number!=self._request_count: return
     
     self.error=True
     self._buffer.clear()
     self.busy=False
  
  def recall(self,chunk,matches,request_number):
     self.finst.add(chunk)
     time=self.latency*math.exp(-chunk.activation)
     if time>self.maximum_time: time=self.maximum_time
     yield time
     if request_number!=self._request_count: return
     self._buffer.set(chunk)
     for a in self.adaptors: a.recalled(chunk)
     self.busy=False
     
  
  def get_activation(self,chunk):
     if not isinstance(chunk,Chunk):
         try:
           chunk=Chunk(chunk,self.sch.bound)
         except AttributeError:
           chunk=Chunk(chunk,None)           
         for c in self.dm:
             if chunk==c:
                 chunk=c
                 break
         else:
             raise Exception('No such chunk found')        
     act=0
     for a in self.adaptors:
        act+=a.activation(chunk)
     if self.record_all_chunks or chunk.record is True:
         self.log[str(chunk)]=act 
     return act   
                
  def add_adaptor(self,a):
     self.adaptors.append(a)

class Finst:
  def __init__(self,parent,size=4,time=3.0):
    self.parent=parent
    self.size=size
    self.time=time
    self.obj=[]
  def contains(self,o):
    return o in self.obj
  def add(self,o):
    if self.size==0: return
    self.obj.append(o)
    if len(self.obj)>self.size:
      self.remove(self.obj[0])
    self.parent.sch.add(self.remove,args=[o],delay=self.time)
  def remove(self,o):
    if o in self.obj: self.obj.remove(o)          
      
    
class MemorySubModule:
  def __init__(self,parent):
    self.parent=parent
    if parent is not None:
      parent.add_adaptor(self)
  def create(self,chunk,**keys):
    pass
  def merge(self,chunk,**keys):
    pass
  def matched(self,chunks):
    pass  
  def activation(self,chunk):
    return 0
  def recalled(self,chunk):
    pass
  def now(self):
    if self.parent is None or not self.parent._is_converted(): return 0
    else: return self.parent.now()

class DMNoise(MemorySubModule):
  def __init__(self,memory,noise=0.3,baseNoise=0.0):
    MemorySubModule.__init__(self,memory)
    self.noise=noise
    self.baseNoise=baseNoise  
  def create(self,chunk,**keys):
    chunk.baseNoise=self.logisticNoise(self.baseNoise)
  def activation(self,chunk):
    return chunk.baseNoise+self.logisticNoise(self.noise)
  def logisticNoise(self,s):
    try:
      x=self.parent.random.random()
    except AttributeError:
      import random
      x=random.random()
    return s*math.log(1.0/x-1.0)
      


class DMBaseLevel(MemorySubModule):
  def __init__(self,memory,decay=0.5,limit=None):
    MemorySubModule.__init__(self,memory)
    self.decay=decay
    self.limit=limit
  
  def create(self,chunk,time=0.0,baselevel=None,**keys):
    chunk.creation=self.now()
    chunk.times=[chunk.creation-time]
    chunk.count=1
    if baselevel is not None and baselevel!='calculate':
      chunk.baselevel=baselevel
    
  def merge(self,chunk,time=0.0,baselevel=None,**keys):
    chunk.times.append(self.now()-time)
    chunk.count+=1
    if self.limit is not None and len(chunk.times)>self.limit:
      if self.limit==0: del chunk.times[:]
      else: chunk.times=chunk.times[-self.limit:]
    if baselevel is not None:
      if baselevel=='calculate': del chunk.baselevel
      else: chunk.baselevel=baselevel
    
  def activation(self,chunk):
    if hasattr(chunk,'baselevel'):
      return chunk.baselevel
    if self.decay==None: 
      return 0
    d=self.decay
    now=self.now()
    t=[now-time for time in chunk.times]
    t=[max(L,0.005) for L in t]
    
    exact=0.0
    approx=0.0
    if len(t)>0:
      exact=sum([math.pow(time,-d) for time in t])
    if self.limit is not None:
      n=chunk.count
      k=len(chunk.times)
      if n>k:
        tn=now-chunk.creation
        if k>0:
          tk=now-chunk.times[0]
        else:
          tk=0.0
        approx=(n-k)/(1-d)*(math.pow(tn,1-d)-math.pow(tk,1-d))/(tn-tk)
    
    B=math.log(exact+approx)
    return B

class DMSpacing(MemorySubModule):
  def __init__(self,memory,decayScale=0.0,decayIntercept=0.5):
    MemorySubModule.__init__(self,memory)
    self.decayScale=decayScale
    self.decayIntercept=decayIntercept  
  
  def create(self,chunk,time=0.0,**keys):
    chunk.times=[(self.actr.time-time,self.decayIntercept)]
        
  def merge(self,chunk,time=0.0,**keys):
    t=self.now()-time
    m=self.activation(chunk)
    d=self.decayScale*math.exp(m)+self.decayIntercept
    chunk.times.append((t,d))
    
  def activation(self,chunk):
    now=self.now()
    total=0.0
    for time,d in chunk.times:
      total+=math.pow(now-time,-d)
    B=math.log(total)
    return B

class DMInhibition(MemorySubModule):
    def __init__(self,memory,decayScale=1.0,timeScale=5.0):
        MemorySubModule.__init__(self,memory)
        self.decayScale=decayScale
        self.timeScale=timeScale
    def create(self,chunk,time=0.0,**keys):
        chunk.mostRecentTime=time
    def merge(self,chunk,time=0.0,**keys):
        chunk.mostRecentTime=self.now()
    def activation(self,chunk):
        tn=self.now()-chunk.mostRecentTime
        I=-math.log(1+math.pow(tn,-self.decayScale)/self.timeScale)
        return I


class DMSalience(MemorySubModule):
  def __init__(self,memory):
    MemorySubModule.__init__(self,memory)
    self.histogram={}
    self.weight={}
    self.memory=memory
  
  def weights(self,**weights):
    self.histogram={}
    self.weight={}
    for k,v in list(weights.items()):
      if k.startswith('_'): k=int(k[1:])
      self.weight[k]=float(v)
      self.histogram[k]={}
    
  def context(self,pattern):
    pat=Pattern(pattern)
    chunks=[x for x in self.memory.dm if pat.match(x) is not None]
    for k,hist in list(self.histogram.items()):
      hist.clear()
    if len(chunks)==0: raise Exception('No chunks match salience pattern: "%s"'%pattern)
    dw=1.0/len(chunks)
    for c in chunks:
      for k,hist in list(self.histogram.items()):
        val=c.get(k,None)
        if val not in hist: hist[val]=dw
        else: hist[val]+=dw
        
  def activation(self,chunk):
    act=0.0
    for k,hist in list(self.histogram.items()):
      val=chunk.get(k,None)
      p=hist[val]
      w=self.weight[k]
      act+=math.log(1.0/p,2)*w
#    if self.log: self.log.act[`chunk`]=act
    return act  
      
      
class DMSpreading(MemorySubModule):
  def __init__(self,memory,*buffers):
    MemorySubModule.__init__(self,memory)
    self.strength=1
    self.buffers=buffers
    self.weight={}
    for b in buffers: 
      self.weight[b]=1
    self.slots={}
    
  def create(self,chunk,**keys):
    for slot in list(chunk.values()):
      if slot in self.slots:
        self.slots[slot].append(chunk)
      else:
        self.slots[slot]=[chunk]
  def activation(self,chunk):
    values=list(chunk.values())
  
    total=0.0
    for b in self.buffers:
     ch=b.chunk 
     if ch is not None:
      w=self.weight[b]
      for key,slot in list(ch.items()):
        if slot in values:
          s=self.strength-math.log(len(self.slots[slot])+1)
          total+=w*s
    return total      
        
class DMFixed(MemorySubModule):
  def __init__(self,memory,default=0):
    MemorySubModule.__init__(self,memory)
    self.default=default
  def create(self,chunk,fixed=None,**keys):
    if fixed is None: fixed=self.default
    chunk.fixed=fixed
  def merge(self,chunk,fixed=0,**keys):
    chunk.fixed+=fixed
  def activation(self,chunk):
    return chunk.fixed


class Associated:
  def __init__(self,a,b):
    self.a=a
    self.b=b

class DMAssociate(MemorySubModule):
  def __init__(self,memory,buffer,weight=1,decay=0.5,limit=None):
    MemorySubModule.__init__(self,memory)
    self._buffer=buffer
    self._mem={}
    self._bl=DMBaseLevel(None,decay=decay,limit=limit)
    self.weight=weight
  def set_association(self,pre,post,baselevel):
    c=self._mem.get((pre,post),None)
    if c==None:
      c=Associated(pre,post)
      self._bl.create(c)
      self._mem[(pre,post)]=c
    c.baselevel=baselevel  
    
  def recalled(self,chunk):
    prechunk=self._buffer.chunk
    if prechunk is None: return
    for pk,pv in list(prechunk.items()):
      for k,v in list(chunk.items()):
        c=self._mem.get((pv,v),None)
        if c==None:
          c=Associated(pv,v)
          self._bl.create(c)
          self._mem[(pv,v)]=c
        else:
          self._bl.merge(c)
  def activation(self,chunk):
    act=0
    prechunk=self._buffer.chunk
    if prechunk is None: return 0
    for pk,pv in list(prechunk.items()):
      for k,v in list(chunk.items()):
        c=self._mem.get((pv,v),None)
        if c is not None:         
          act+=self._bl.activation(c)
    return act*self.weight
        
        
        
        
    
          
    
class Partial:
  def __init__(self,memory,strength=1.0,limit=-1.0):
    self._memory=memory
    self.limit=limit
    self.strength=strength
    self.sims={}
    memory.partials.append(self)
    
  def similarity(self,a,b,value):
    #import pdb; pdb.set_trace()
    self.sims[a,b]=value
    self.sims[b,a]=value
  
  def request(self,pattern):
    self._memory.request(pattern,partial=self)
    
  def match(self,key,a,b):
    m=self.sims.get((a,b),self.limit)
    p=self.strength
    return p*m
    
      
class BlendingMemory(Memory):
  def recall(self,chunk,matches,request_number):
    keys=list(sorted(chunk.keys()))
    if not hasattr(chunk,'_blend_keys'):
      bk=[]
      for k in keys:
        try:
          v=float(chunk[k])
          bk.append(k)
        except:
          pass
      chunk._blend_keys=tuple(bk)
    bk=chunk._blend_keys
    if len(bk)>0:
      a=chunk.activation
      chunk=Chunk(chunk)
      chunk.activation=a
      
      for b in bk: chunk[b]=0
      total_activation=0
      for m in matches:
        m.exp_activation=math.exp(m.activation)
        k=list(sorted(m.keys()))
        if k==keys:
          blend={}
          try:            
            for b in bk:
              blend[b]=float(m[b])
          except ValueError:
            continue
          for b in bk:
            chunk[b]+=blend[b]*m.exp_activation
          total_activation+=m.exp_activation
      for b in bk:
        chunk[b]/=total_activation
    for m in Memory.recall(self,chunk,matches,request_number):
      yield m
      
    
  
  




from python_actr.pattern import Pattern,PatternException

class Obj:
  def __init__(self,**keys):
    for k,v in list(keys.items()):
      self.__dict__[k]=v
      
def makePattern(text):
  return Pattern(dict(self=text))
      

def test_exact():
    obj={'self':Obj(a=1,b=2,c=3)}
    p=Pattern(dict(self='a:1 b:2 c:3'))
    assert p.match(obj) == {}
    obj['self'].c=2
    assert p.match(obj) == None
    assert Pattern(dict(self='a:1 a:2')).match(obj) == None
    assert p.match({'self':None}) == None
    assert Pattern(dict(self=None)).match({'self':None}) == {}
    
    
def test_notexact():
    obj={'self':Obj(a=1,b=2,c=3)}
    p=makePattern('a:!2 b:2')
    assert p.match(obj) == {}
    obj['self'].a=2
    assert p.match(obj) == None
    
def test_vars():
    obj={'self':Obj(a=1,b=2,c=1)}
    
    assert makePattern('a:?x').match(obj) == {'x':'1'}
    assert makePattern('a:?x c:?x').match(obj) == {'x':'1'}
    assert makePattern('a:?x b:?x').match(obj) == None
    assert makePattern('a:?x b:?y').match(obj) == {'x':'1','y':'2'}
    assert makePattern('a:?x b:!?x').match(obj) == {'x':'1'}
    assert makePattern('a:?x c:!?x').match(obj) == None
    
def test_unnamed_slot():
    obj={'self':[1,2,1]}

    assert makePattern('?x').match(obj) == {'x':'1'}
    assert makePattern('?x ? ?x').match(obj) == {'x':'1'}
    assert makePattern('?x ?x').match(obj) == None
    assert makePattern('?x ?y').match(obj) == {'x':'1','y':'2'}
    assert makePattern('?x !?x').match(obj) == {'x':'1'}
    assert makePattern('?x ? !?x').match(obj) == None
        
    assert makePattern('?x 1:?y').match(obj) == {'x':'1','y':'2'}
    assert makePattern('?x 2:?x').match(obj) == {'x':'1'}
    assert PatternException,makePattern == ('a:1 2')
  
    
def test_str():
    obj={'self':dict(a=True,b=False,c=None)}
    
    assert makePattern('a:True').match(obj) == {}
    assert makePattern('b:False').match(obj) == {}
    assert makePattern('c:None').match(obj) == {}
    
    assert makePattern('a:False').match(obj) == None
    assert makePattern('b:None').match(obj) == None
    assert makePattern('c:False').match(obj) == None
    
def test_group():
    obj={'a':dict(a=1,b=None),'c':dict(d=2,e=1)}
    
    assert Pattern(dict(c='d:2 e:1')).match(obj) == {}
    assert Pattern(dict(a='a:?x',c='d:2 e:?x')).match(obj) == {'x':'1'}
    assert Pattern(dict(a='b:!?x',c='d:!?x e:?x')).match(obj) == {'x':'1'}
    assert Pattern(dict(a='a:!?x b:!?x',c='d:!?x e:?x')).match(obj) == None
  
def test_lambda():
    obj={'self':Obj(a=1,b=2,c=1)}

    assert makePattern(lambda x,b: x.a+x.b+x.c==4).match(obj) == {}
    assert makePattern(lambda x,b: x.a+x.b+x.c==3).match(obj) == None      
    assert makePattern(['a:?z',lambda x,b: x.a+x.b+x.c-int(b['z'])==3]).match(obj) == {'z':'1'}
    assert makePattern(['a:?z',lambda x,b: x.a+x.b+x.c+int(b['z'])==3]).match(obj) == None


import python_actr

from collections import UserDict

class Chunk(UserDict):
  def __init__(self,contents,bound=None):
    UserDict.__init__(self)
    if isinstance(contents,Chunk):
      self.update(contents)
    elif isinstance(contents,str):
      for i,x in enumerate(contents.split()):
        if ':' in x:
          i,x=x.split(':',1)
        if x.startswith('?'):
          key=x[1:]
          x=bound[key]
        self[i]=x
    elif hasattr(contents,'__dict__'):  
      for k,v in list(contents.__dict__.items()):
        if type(v) in [str,float,int,bool]:
          self[k]=v
    else:
      try:
        self.update(contents)
      except:  
        raise Exception('Unknown contents for chunk:',contents)      
  def __repr__(self):
    #
    r=[]
    keys=list(self.keys())
    i=0
    while i in keys:
      r.append('%s'%self[i])      
      keys.remove(i)
      i+=1
    keys.sort()
    for k in keys:
      if k[0]!='_':
        r.append('%s:%s'%(k,self[k]))
    return ' '.join(r)  
      

class Buffer(python_actr.Model):
  def __init__(self):
    self.chunk=None
  def set(self,chunk):
    try:
      self.chunk=Chunk(chunk,self.sch.bound)
    except AttributeError:
      self.chunk=Chunk(chunk,{})  
  def modify(self,**args):
    for k,v in list(args.items()):
      if k.startswith('_'): k=int(k[1:])
      if k not in self.chunk:
        raise Exception('No slot "%s" to modify to "%s"'%(k,v))
      self.chunk[k]=v
      self.chunk=self.chunk
  def __getitem__(self,key):
    return self.chunk[key]
  def clear(self):
    self.chunk=None  
  def __eq__(self,other):
    return self.chunk==other
  def __hash__(self):
    return id(self)
  def __len__(self):
    if self.chunk is None: return 0
    return len(self.chunk)
  def isEmpty(self):
    return len(self)==0



import unittest

import python_actr
from python_actr import *


class Basic(ACTR):
    def init():
        self.x=1
        
    def p1(self='x:1'):
        self.x=2    


class UtilLearn(ACTR):
    x=1
    c1=0
    c2=0
    def p1(self='x:1'):
        self.success()
        self.c1+=1
    def p2(self='x:1'):
        self.failure()
        self.c2+=1


class UtilSetting(ACTR):
    x=0
    def p1(self='x:0',utility=0.3):
        self.x=1
    def p2(self='x:0',utility=0.6):
        self.x=2
    



class TestACTRProduction(unittest.TestCase):
    def test_basic(self):
        p=Basic()
#        ccm.log_everything(p)
        p.run()
        self.assertEqual(p.x,2)
    
    def test_utillearn(self):
        for Rule in PMPGC,PMNew,PMPGCSuccessWeighted,PMPGCMixedWeighted,PMTD,PMNew:
          p=UtilLearn()
          p.pm=Rule()
#          ccm.log_everything(p)
          p.run(limit=1)
          self.assertTrue(p.c1 in [17,18,19])
  
    
    def test_setting_utility(self):
        p=UtilSetting()
        p.run()
        self.assertEqual(p.x,2)
    
    
    


if __name__ == '__main__':
  unittest.main()     





from . import model
import inspect
import re
from . import pattern

try:
    Set=set
except:
    from sets import Set

class ProductionException(Exception):
    pass

class Production:
    def __init__(self,system,name,func):
        self.system=system
        self.name=name
        self.base_utility=0
        spec = inspect.getfullargspec(func)
        self.keys = spec.args
        patterns={}
        for i,name in enumerate(spec.args[:]):
          if name=='utility': 
            self.base_utility = spec.defaults[i]
            del spec.args[i]
          else:
            patterns[name] = spec.defaults[i]
        self.pattern_specs=patterns
        self.pattern=pattern.Pattern(patterns)
        self.bound=None
        
        self.original_func=func
        code=inspect.getsource(func)
        m=re.match(r'[^(]+\([^(]*\):',code)
        self.code=code[m.end():]
        code='if True:'+code[m.end():]
        self.func=compile(code,'<production-%s>'%self.name,'exec')
        
    def match(self,obj):
        b=self.pattern.match(obj)
        if b is None: return False
        self.bound=b
        return True
    
    def fire(self,context):
        self.system.sch.bound=self.bound
        exec(self.func, context,self.bound)
            
      
class ProductionSystem(model.Model):
    production_time=0.05
    production_match_delay=0
    _auto_run_start=False
    def _convert_info(self,objects,methods):
        self._productions=[]
        self._initializers=[]
        self._keys_used=Set()
        for k,v in list(methods.items()):
            spec = inspect.getfullargspec(v)
            if spec.varargs is None and spec.varkw is None:
              if spec.defaults is None and len(spec.args) == 0:
                p=Production(self,k,v)
                self._initializers.append(p)
              if spec.defaults is not None and spec.args is not None and len(spec.args) == len(spec.defaults):
                p=Production(self,k,v)
                self._keys_used.update(p.keys)
                self._productions.append(p)
        self.sch.add(self._process_productions)
    
    def _calc_context(self):
        context={}
        keys=Set(self._keys_used)
        if 'self' in keys: keys.remove('self')
        if len(keys)==0: top=self
        m=self
        while m is not None:
            for k,v in list(m.__dict__.items()):
                if k not in context and k[0]!='_' and k!='parent' and isinstance(v,object) and not isinstance(v,model.MethodWrapper):
                    context[k]=v
                    if k in keys: 
                        keys.remove(k)
                        if len(keys)==0: top=m
            if 'top' in keys: top=m            
            m=m.parent
            
        if 'top' in keys: keys.remove('top') 
        if len(keys)>0:
            raise ProductionException("Production is matching on an unknown module '%s'"%(keys))

        #TODO: rethink this. top should refer to the highest level in the tree.
        # but it would be nice if we had something to refer to the highest level
        # that could cause a matching change, to optimize the yield._top.changes
        # call.  This was the original idea and may impact the logic calculating
        # top above.  However, any fix to this should make sure not to break the
        # rock paper scissors tutorials.
        while hasattr(top,'parent') and top.parent is not None: top=top.parent
        
        context['self']=self
        context['top']=top
        self._top=top
        self._context=context
    
    def _process_productions(self):
        self._calc_context()          
        for i in self._initializers:
            i.fire(self._context)
        while True:
          if self.production_match_delay>0: yield self.production_match_delay  
          match=[p for p in self._productions if p.match(self._context)]
          if len(match)==0:
            yield self._top.changes
          else:
            choice=self.random.choice(match)
            self.log.production=choice.name
            yield self.production_time-self.production_match_delay
            self.log.production=None
            choice.fire(self._context)
            
        
                          
            
            
        




import python_actr

class Timer(python_actr.Model):
    def __init__(self,startpulse=0.011,a=1.1,b=0.015):
        self.startpulse=startpulse
        self.a=a
        self.b=b
        
        self.count=0
        self.t=self.startpulse    
        self.start()
    def reset(self):
        yield self.startpulse
        self.count=0
        self.t=self.startpulse    
    def start(self):
        while True:
            self.t=self.random.gauss(self.a*self.startpulse,self.a*self.startpulse*self.b)
            self.count+=1
            yield self.t,reset.finished




from pathlib import Path
from setuptools import setup

# read the contents of the README file
cwd = Path(__file__).parent
long_description = (cwd / "README.md").read_text()

setup(
    name="actr",
    # version is the actr pip package (this is different from the python_actr version)
    version="1.0.0",
    description="Python implementation of the ACT-R cognitive architecture",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Carleton Cognitive Modelling Lab",
    maintainer="Andy Maloney (fork)",
    url="https://github.com/asmaloney/python_actr",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
    ],
    license="MIT",
    keywords=[
        "ACT-R",
        "cognitive architecture",
        "cognitive modelling",
        "cognitive science",
        "CogSci",
    ],
    packages=[
        "python_actr",
        "python_actr.display",
        "python_actr.actr",
        "python_actr.ui",
    ],
    python_requires=">=3.0",
)



"""
PyTag
Version: 2005-06-11
Author: Terry Stewart terry.stewart@gmail.com http://terrystewart.ca

This software is released under the GNU General Public License.
See http://www.gnu.org/copyleft/gpl.html for more details.

This is a utility for easily generating valid XHTML from within
a Python program.  It is inspired by Nevow's <http://nevow.com> 'Stan'
module, which is pretty much exactly the same as this one.  I just
wrote my own version because I felt like it.

Here is a quick example of what you can do:

from pytag import T
print T.html[
  T.head[T.title["This is the title"]],
  T.body[
    T.h1["Heading"],
    T.hr,
    T.p(align='center')[
      "Hello and welcome to my ",
      T.em["nifty"],
      " web page."
      ]
    ]
  ]

As you can see, this is a rather odd (ab)use of the [] and () syntax.
You create tags by saying T.tagname (using any tagname you want).

You set the attributes of the tag by "calling" it.
   T.tag(name=value, othername=othervalue)
You set the contents of the tag using the square brackets.
   T.tag['some stuff',T.othertag,'can go here']
You can do them both at the same time, so you can define an
html link like this:
   T.a(href='page.html')['click here']
or
   T.a['click here'](href='page.html')

You can even do this to simplify the syntax for common tags:

from pytag import *
print html[
  head[title["This is the title"]],
  body[
    h1["Heading"],
    hr,
    p(align='center')[
      "Hello and welcome to my ",
      em["nifty"],
      " web page."
      ]
    ]
  ]


If you add things more than once, they get appended together

  myList=T.ol
  for i in range(10):
    mylist[T.li['this is item #%d'%i]]
  print myList

You can also use the Python list comletion syntax to do something
like that

  print T.ol[[T.li['this is item #%d'%i for i in range(10)]]]

(note the currently-required double-square brackets in that example.
It is possible that that will become optional when list completions
become more general in newer versions of Python)

One other capability is to create template tags.  These are just like
normal tags, but when you add things to them they create a new
instance of the tag, rather than accumulating.

  myP=T.p(align='center',size='+1')
  myP.lock()

  print T.body[
    T.h1['Header'],
    myP['Here is paragraph 1'],
    myP['Here is paragraph 2'],
    ]

For further documentation, see http://terrystewart.ca/pytag.html

**** History
- 2005-06-11: Initial Release
"""



import copy

def _flatten(args):
  for arg in args:
    if type(arg) in (type(()),type([])):
      for x in arg:
        for f in _flatten(x): yield f
    else: yield arg

class Tag:
  def __init__(self,name):
    self.name=name
    self.attr={}
    self.content=[]
    self.locked=False
  def lock(self):
    self.locked=True
    return self
  def __getitem__(self,content):
    if self.locked:
      self=copy.deepcopy(self)
      self.locked=False
    if type(content) not in (type(()),type([])):
      self.content.append(content)
    else:
      self.content.extend(_flatten(content))
    return self
  def __call__(self,**attr):
    if self.locked: 
      self=copy.deepcopy(self)
      self.locked=False
    for k,v in list(attr.items()):
      v=str(v)
      v='"%s"'%v
      self.attr[k]=v
    return self
  def __str__(self):
    attr=''
    if len(self.attr)>0:
      attr=' '+' '.join(['%s=%s'%kv for kv in list(self.attr.items())])
    if len(self.content)==0:
      return '<%s%s />'%(self.name,attr)
    else:
      content=''.join([str(x) for x in self.content])
      return '<%s%s>%s</%s>'%(self.name,attr,content,self.name)

class _TagMaker:
  def __getattr__(self,key):
    return Tag(key)
T=_TagMaker()


_tagList="""html head title body p em strong ol li ul blockquote br
center cite code dd dt dl h1 h2 h3 h4 h5 h6 hr img meta small big a
form input option select optgroup textarea style
sub sup tt
table tr td th
""".split()
for t in _tagList:
  exec('%s=T.%s;%s.lock()'%(t,t,t))


def _pytagtest():
  print(p['Hello world'])

  print(body[T.p['Hello']])

  print(html[
    head[T.title["This is the title"]],
    body[
      h1["Heading"],
      p(align='center')[
        "Hello and welcome to my ",
        em["nifty"],
        " web page."
        ],
      br,
      "Click ",a(href="page.html")['here']," to go somewhere else."
      ]
    ])

  print(ol[[li[x] for x in range(10)]])

  cp=p(align='center')
  cp.lock()
  print(p[cp(a='b')['p 1'],cp['p 2']])



if __name__=='__main__':
  _pytagtest()




import heapq
import copy

from . import logger

class Trigger:
    def __init__(self,name=''):
        self.name=name
    def __str__(self):
        return '<Trigger "%s">'%self.name    

class Event:
  generator=False
  def __init__(self,func,time,args=[],keys={},priority=0):    
    self.name=getattr(func,'func_name',None)

    try:
       code=func.__code__
    except AttributeError:
       try:
          code=func.__call__.__func__.__code__   
       except:
          code=None   
    if code and code.co_flags&0x20==0x20:    # check for generator
        func=func(*args,**keys).__next__
        args=[]
        keys={}
        self.generator=True

    self.func=func
    self.args=args
    self.keys=keys
    self.time=time
    self.priority=priority
    self.group=()
    self.cancelled=False
    self.parent=None
  #def __cmp__(self,other):
    #return cmp((self.time,-self.priority),(other.time,-other.priority))
  
  def __lt__(self,other):
    #print(self.time,-self.priority,'other',other.time,-other.priority)
    return (self.time,-self.priority) < (other.time,-other.priority)
    #return cmp((self.time,-self.priority),(other.time,-other.priority))

  def __le__(self, other):
    return cmp((self.time,-self.priority),(other.time,-other.priority))

  def __eq__(self,other):
    return cmp((self.time,-self.priority),(other.time,-other.priority))

  def __ge__(self, other):
    return cmp((self.time,-self.priority),(other.time,-other.priority))
  
  def __gt__(self,other):
    return cmp((self.time,-self.priority),(other.time,-other.priority))

  def __ne__(self, other):
    return cmp((self.time,-self.priority),(other.time,-other.priority))  

  def __repr__(self):
    return '<%s %x %5.3f>'%(self.name,id(self.func),self.time)
    
class SchedulerError(Exception):
    pass

class Scheduler:
    def __init__(self):
        self.queue=[]
        self.to_be_added=[]
        self.triggers={}
        self.time=0.0
        self.stop_flag=False
        self.log=logger.log_proxy
    def extend(self,other):
        for k,v in list(other.triggers.items()):    
            if k not in self.triggers:
                self.triggers[k]=v
            else:
                self.triggers[k].extend(v)
        if len(other.queue)>0:
            self.queue.extend(other.queue)
            heapq.heapify(self.queue)            
    def trigger(self,key,priority=None):
        if 'OpenGL' in key.name:
          print(key.name)
          print(key in self.triggers)
        if key in self.triggers:
            for event in self.triggers[key]:
                event.time=self.time
                if priority is not None:
                   event.priority=priority
                self.add_event(event)
            del self.triggers[key][:]
    def add_event(self,event):
        heapq.heappush(self.queue,event)
    def add(self,func,delay=0,args=[],keys={},priority=0,thread_safe=False):
        if thread_safe:
          self.to_be_added.append((func,delay,args,keys,priority))
        else:  
          ev=Event(func,self.time+delay,args=args,keys=keys,priority=priority)
          self.add_event(ev)
          return ev
        
    def run(self):
        self.stop_flag=False
        while not self.stop_flag and len(self.queue)>0:
            next=self.queue[0].time
            if next>self.time:
                self.time=next
                self.log.time=next
            self.do_event(heapq.heappop(self.queue))    
            while self.to_be_added:
              self.add(*self.to_be_added.pop())
        
    def handle_result(self,result,event):
        if isinstance(result,(int,float)):
            event.time=self.time+result
            self.add_event(event)
        elif isinstance(result,dict):
            event.time=self.time+result.get('delay',0)
            event.priority=result.get('priority',event.priority)
            self.add_event(event)
        elif isinstance(result,(str,Trigger)):
            event.time=None
            if result not in self.triggers:
                self.triggers[result]=[event]
            else:
                self.triggers[result].append(event)
        elif isinstance(result,(list,tuple)):
            events=[copy.copy(event) for r in result]
            for e in events: e.group=events
            for i,r in enumerate(result):
                self.handle_result(r,events[i])
        elif result is None:
            if event.parent is not None:
                event.parent.time=self.time
                self.add_event(event.parent)
        elif isinstance(result,Event):
            if result.generator and event.generator:
                result.parent=event
        elif hasattr(result,'default_trigger'):
          self.handle_result(result.default_trigger,event)        
        else:
            raise SchedulerError("Incorrect 'yield': %s"%(result))
                
            
            
    def do_event(self,event):
        assert self.time==event.time

        if event.cancelled: return
        for e in event.group: e.cancelled=True
        event.cancelled=False

        try:
          result=event.func(*event.args,**event.keys)
        except StopIteration:
          result=None
          
        self.handle_result(result,event)
        
                               
            
            
          
        
    
    def stop(self):    
      self.stop_flag=True  



import python_actr
import math
import collections

__all__=['ProceduralSubModule','PMNoise','PMPGC',
         'PMPGCSuccessWeighted','PMPGCMixedWeighted',
         'PMTD','PMNew']

class ProceduralSubModule(python_actr.Model):
  def start(self):
    self.parent.add_adaptor(self)
  def create(self,prod):
    pass
  def firing(self,prod):
    pass
  def selecting(self,prod):
    pass
  def reward(self,prod):
    pass
  def utility(self,prod):
    return 0      
  def below_threshold(self):
    pass  
       

class PMNoise(ProceduralSubModule):
  def __init__(self,noise=0,baseNoise=0.0):
    self.noise=noise
    self.baseNoise=baseNoise
  def create(self,prod,parents=None):
    prod.baseNoise=self.logisticNoise(self.baseNoise)
  def utility(self,prod):
    return prod.baseNoise+self.logisticNoise(self.noise)
  def logisticNoise(self,s):
    x=self.random.random()
    return s*math.log(1.0/x -1.0)
    
class PMPGC(ProceduralSubModule):
  def __init__(self,goal=20):
    self.history=[]
    self.goal=goal
    self._clearFlag=False
  def create(self,prod,parents=None):
    #todo: modify this to do the PGC estimate based on parents if there are any
    prod.successes=1
    prod.failures=0
    prod.time=self.parent.production_time
    if isinstance(prod.time, collections.Callable): prod.time=prod.time()
    prod.lock_pgc=False
  def selecting(self,prod):
    if self._clearFlag:
      del self.history[:]
      self._clearFlag=False
    self.history.append((prod,self.now()))  
  def reward(self,value):
    now=self.now()
    for p,t in self.history:
        if not p.lock_pgc:
          dt=now-t
          p.time+=dt
          if value>=0: p.successes+=value
          else: p.failures-=value
    self._clearFlag=True  
  def utility(self,prod):
    p=prod.successes/(prod.successes+prod.failures)
    c=prod.time/(prod.successes+prod.failures)
    g=self.goal
    return p*g-c
  def set(self,prod,successes=None,failures=None,time=None,lock=None):
    if not isinstance(prod,python_actr.Production):
        prod=self.parent._productions[prod]
    if successes is not None: prod.successes=successes
    if failures is not None: prod.failures=failures
    if time is not None: prod.time=time
    if lock is not None: prod.lock_pgc=lock
    
    
      

class PMPGCSuccessWeighted(PMPGC):
  def utility(self,prod):
    p=1.0
    c=prod.time/(prod.successes)
    g=self.goal
    return p*g-c
class PMPGCMixedWeighted(PMPGC):
  def utility(self,prod):
    p=prod.successes/(prod.successes+prod.failures)
    c=prod.time/(prod.successes)
    g=self.goal
    return p*g-c
  

# From (Fu & Anderson, 2004)
class PMTD(ProceduralSubModule):
    def __init__(self,alpha=0.1,discount=1,cost=0.05):
        self.alpha=alpha
        self.discount=discount
        self.last_prod=None
        self.last_time=None
        self.this_reward=0
        self.cost=cost
    def create(self,prod,parents=None):
        prod.td_u=0
        prod.cost=self.cost
    def firing(self,prod):
        if self.last_prod is not None:
            d=self.now()-self.last_time
            k=self.discount
            r=self.this_reward-self.last_prod.cost+prod.td_u/(1+k*d)
            td=r-self.last_prod.td_u
            self.last_prod.td_u+=self.alpha*td
        self.last_prod=prod
        self.last_time=self.now()
        self.this_reward=0

    def reward(self,value):
        self.this_reward+=value

    def utility(self,prod):
        return prod.td_u



       
                       
class PMNew(ProceduralSubModule):
  def __init__(self,alpha=0.2):
    self.history=[]
    self.alpha=alpha
    self.clearFlag=False
  def create(self,prod,parents=None):
    prod.util=0
  def selecting(self,prod):
    if self.clearFlag:
      del self.history[:]
      self.clearFlag=False
    self.history.append((prod,self.now()))  
  def reward(self,value):
    now=self.now()
    for p,t in self.history:
      dt=now-t
      r=value-dt
      p.util+=self.alpha*(r-p.util)
    self.clearFlag=True  
  def utility(self,prod):
    return prod.util
            



import python_actr

class Motor(python_actr.Model):
  def __init__(self):
    python_actr.Model.__init__(self)
    self.busy=False

  def press(self,key):
    if self.busy: return
    self.busy=True
    self.log._='Pressing key preparation'
    yield 0.25
    self.log._='Pressing key preparation complete'
    yield 0.05
    self.log._='Pressing key initiation complete'
    yield 0.1
    self.log._='Actually pressing the key'
    self.parent.parent.key_pressed(key)
    self.log._='Finishing movement'
    yield 0.15
    self.log._='Finished movement'
    self.busy=False




import unittest

from actr import *

class Basic(ACTR):
    buffer1=Buffer()
    def init():
        buffer1.set('hello world')
        
    def bye(buffer1='hello world'):
        buffer1.set('goodbye world')
        
    def quit(buffer1='goodbye world'):
        self.stop()

class Vars(ACTR):
    buffer1=Buffer()
    def init():
        buffer1.set('hello world')
        
    def bye(buffer1='hello ?text'):
        buffer1.set('goodbye ?text')
        
    def quit(buffer1='goodbye ?'):
        self.stop()


class TestACTRProduction(unittest.TestCase):
    def test_basic(self):
        p=Basic()
#        python_actr.log_everything(p)
        p.run()
        self.assertEqual(p.buffer1[0],'goodbye')
        self.assertEqual(p.now(),0.1)
            
    def test_vars(self):
        p=Vars()
#        python_actr.log_everything(p)
        p.run()
        self.assertEqual(p.buffer1[0],'goodbye')
        self.assertEqual(p.now(),0.1)
    
    def test_chunk_setting(self):
        from python_actr.actr.buffer import Chunk
        c=Chunk(dict(a=1,b=2))
        self.assertEqual(c['a'],1)
        self.assertEqual(c['b'],2)
      
    
    


if __name__ == '__main__':
  unittest.main()     



import python_actr
import unittest

class SubSubModel(python_actr.Model):
  def start(self):
    self.value=0
    yield self.random.random()
    self.value+=1


class SubModel(python_actr.Model):
  def start(self):
    self.value=0
    yield self.random.random()
    self.value+=1
  subsub1=SubSubModel()
  subsub2=SubSubModel()

class BaseModel(python_actr.Model):
  def start(self):
    self.value=0
    yield 1
    self.value+=1
  
  sub1=SubModel()
  sub2=SubModel()

class ModuleA(python_actr.Model):
  def __init__(self,other):
    self.other=other
  def start(self):
    yield 0.5
    self.other.a='a'
    yield 0.5
    self.other.store(1)

class ModuleB(python_actr.Model):
  def store(self,value):
    self.value=value

class Joined(python_actr.Model):
  b=ModuleB()
  a=ModuleA(b)
 

class TestNested(unittest.TestCase):
  def test_nested(self):
    m=BaseModel()
#    python_actr.log_everything(m)
    m.run()
    self.assertEqual(m.value,1)
    self.assertEqual(m.sub1.value,1)
    self.assertEqual(m.sub2.value,1)
    self.assertEqual(m.sub1.subsub1.value,1)
    self.assertEqual(m.sub2.subsub1.value,1)
  
  def test_joined(self):
    m=Joined()  
#    python_actr.log_everything(m)
    m.run()
    self.assertEqual(m.b.a,'a')
    self.assertEqual(m.b.value,1)



if __name__ == '__main__':
    unittest.main()    




from python_actr.actr.pm import ProceduralSubModule
from python_actr.production import Production
from python_actr.pattern import Pattern

class CompiledProduction(Production):
    def __init__(self,pre,post,keep,retrieve,pre_bound,post_bound):
        self.name='%s-%s-%d'%(pre.name,post.name,id(self))
        self.system=pre.system
        self.base_utility=0
        self.bound=None

        code1=[]
        for k,v in list(pre_bound.items()):
            code1.append(' %s=%s'%(k,repr(v)))
        code1.append(' if True:  # compiled from %s'%pre.name)    
        for line in pre.code.splitlines():
            for k in keep:
                if line.strip().startswith(k):
                    code1.append('  '+line)
                    break

        code1='\n'.join(code1)
        for k,v in list(pre_bound.items()):
            code1=code1.replace('?'+k,v)


            
        code2=[]    
        for k,v in list(post_bound.items()):
            code2.append(' %s=%s'%(k,repr(v)))
        code2.append(' if True:  # compiled from %s'%post.name)
        for line in post.code.splitlines():
            if len(line.strip())>0:
                code2.append('  '+line)

        code2='\n'.join(code2)
        for k,v in list(post_bound.items()):
            code2=code2.replace('?'+k,v)

        self.code='if True:\n%s\n%s'%(code1,code2)

        self.func=compile(self.code,'<production-%s>'%self.name,'exec')

        
        keys=list(pre.keys)
        patterns={}
        for buf,pat in list(pre.pattern_specs.items()):
            for k,v in list(pre_bound.items()):
                pat=pat.replace('?'+k,v)
            patterns[buf]=pat
            
        for m in post.keys:
            if m==retrieve: pass
            elif m not in keys:
                keys.append(m)
                pat=post.pattern_specs[m]
                for k,v in list(post_bound.items()):
                    pat=pat.replace('?'+k,v)
                patterns[buf]=pat
                
        self.keys=keys
        self.pattern_specs=patterns
        self.pattern=Pattern(patterns)
       
      

        

class PMCompile(ProceduralSubModule):
    def __init__(self,keep,request,retrieve):
        if not isinstance(keep,(list,tuple)): keep=(keep,)
        self.keep=keep
        self.request=request
        self.retrieve=retrieve
        self.log=None
        self.pre=[]
        self.post=[]
        self._previous=None
        self.compiled={}
    def create(self,prod,parents=None):
        if self.retrieve in prod.keys:
            for m in prod.keys:
                if m not in self.keep and m!=self.retrieve:
                    break
            else:
                self.post.append(prod)

        code=prod.code
        good=False
        for line in code.splitlines():
            line=line.strip()
            if len(line)>0:
                keep=False
                if line.startswith(self.request):
                    good=True
                    continue
                for k in self.keep:
                    if line.startswith(k):
                        keep=True
                        break
                if keep:
                    continue
                else:
                    good=False
                    break
        if good:
            self.pre.append(prod)
    def firing(self,prod):
        if self._previous is not None and prod in self.post:
            self.compile(self._previous,self._previousBound,prod,prod.bound)
        self._previous=None
        if prod in self.pre:
            self._previous=prod
            self._previousBound=dict(prod.bound)
            
    def compile(self,pre,pre_bound,post,post_bound):
        id=(pre,post,tuple(sorted(pre_bound.items())),tuple(sorted(post_bound.items())))
        p=self.compiled.get(id,None)
        if p is None:
            p=CompiledProduction(pre,post,self.keep,self.retrieve,pre_bound,post_bound)
            self.compiled[id]=p
            #print p.name
            #print p.code

            for a in self.parent._adaptors:
                a.create(p,parents=[pre,post])
            self.parent._productions.append(p)
        




import python_actr
from python_actr.pattern import Pattern

class SOSVision(python_actr.Model):
  def __init__(self,buffer,delay=0.0,log=None,delay_sd=None):
    self._buffer=buffer
    self.delay=delay
    self.delay_sd=delay_sd
    self.error=False
    self.busy=False
  
  def request(self,pattern=''):
    if self.busy: return

    matcher=Pattern(pattern)
      
    self.error=False
    r=[]
    for obj in self.parent.parent.get_children():
      if matcher.match(obj)!=None:
        if not hasattr(obj,'salience') and not hasattr(obj,'visible'):
          continue
        
        if hasattr(obj,'salience'):
          if self.random.random()>obj.salience:
            continue
        if hasattr(obj,'visible'):
          if obj.visible==False:
            continue
        if hasattr(obj,'value'):
          if obj.value==None:
            continue
        r.append(obj)

    self.busy=True
    d=self.delay
    if self.delay_sd is not None:
        d=max(0,self.random.gauss(d,self.delay_sd))
    yield d
    self.busy=False

    if len(r)==0:
      self._buffer.clear()
      self.error=True
    else:
      obj=self.random.choice(r)
      if obj not in self.parent.parent.get_children():
        self._buffer.clear()
        self.error=True
      elif hasattr(obj,'visible') and obj.visible==False:
        self._buffer.clear()
        self.error=True
      else:
        self._buffer.set(obj)
      
      
      
        
